/*
*	SCCS:   @(#)	INC/FPCOMP.h		4.1 (01/21/97)  VSXgen release 1.4
*
*	UniSoft Ltd., London, England
*
* (C) Copyright 1991 X/Open Company Limited
*
* All rights reserved.  No part of this source code may be reproduced,
* stored in a retrieval system, or transmitted, in any form or by any
* means, electronic, mechanical, photocopying, recording or otherwise,
* except as stated in the end-user licence agreement, without the prior
* permission of the copyright owners.
*
* X/Open and the 'X' symbol are trademarks of X/Open Company Limited in
* the UK and other countries.
*/


/***********************************************************************

NAME:		FPCOMP.h - routines for floating point comparisons, etc.

PROJECT:	VSX (X/OPEN Validation Suite)

MODULE CONTAINS:
		mlmachar(), fcomp(), dcomp(), ldcomp()

SYNOPSIS:
		void mlmachar(void);
		int fcomp(float val1, float val2);
		int dcomp(double val1, double val2);
		int ldcomp(long double val1, long double val2);

DESCRIPTION:
		Mlmachar() calculates the following external constants
		which are used in the mathematical library tests:-
		    double epspos - smallest positive number which changes 1.0
		    double epsneg - smallest negative number which changes 1.0
		    double maxdble - largest double precision number
		    double mindble - smallest double precision number
		    int expbase - exponent base (normally 2)
		    int dexplen - exponent length in a double
		    int dsignif - number of significant digits in a double
		    int fsignif - number of significant digits in a float
		    int dmaxexp - largest exponent for a double
		    int dminexp - smallest exponent for a double
		
		fcomp(), dcomp() and ldcomp() compare two values of type
		float, double and long double respectively.  They return 1
		if the values are approximately equal, otherwise 0.

REQUIRES PRIOR INCLUSION OF:
		"./std.h" in header tests, if ldcomp() is to be used.

RELATED DOCUMENTS:
AUTHOR:		Andrew Twigger, UniSoft Limited.
DATE CREATED:	Wed Nov 26 15:04:11 BST 1986
MODIFICATIONS:
		Added ldcomp 
		Stuart Boutell, UniSoft Limited.
		Tue Oct  9 15:50:11 BST 1990

		Geoff Clare, 1 Feb 1991
			Merged various versions of these routines into
			one location (here).  Access to them from ".os"
			testsets is now provided via genlib/fpcomp.c,
			which just #includes this file.
		
		Geoff Clare, 13 Mar 1991
			Set tet_nosigreset after ignoring SIGFPE.
		
		Geoff Clare, 8 May 1991
			Add another st() to mindble adjustment.
		
		Geoff Clare, 30 June 1994
			Do not set SIGFPE to be ignored.
		
		Geoff Clare, 6-19 July 1995
			Calculate dmaxexp and maxdble from scratch
			instead of using dminexp and mindble.
			Guard against overflow of d1/d2 in {f,d,ld}comp().

************************************************************************/

#include <stdio.h>
#include <signal.h>
#include <setjmp.h>
#include <errno.h>

#if !__STDC__
#define volatile
#endif

double epspos;
double epsneg;
double maxdble;
double mindble;
int    expbase;
int    dexplen;
int    dsignif;
int    fsignif;
int    dmaxexp;
int    dminexp;
int	hwfactor = 1;

static char normbits[sizeof(double)];
static struct sigaction	old_sigact;
static sigjmp_buf	fpe_jmp;

/* ARGSUSED */
static void
sig_catch(sig)
int sig;
{
	siglongjmp(fpe_jmp, 1);
}

static int
set_signal(handler)
void (*handler)();
{
	struct	sigaction new;

	new.sa_handler = handler;
	new.sa_flags = 0;
	(void)	sigemptyset(&new.sa_mask);

	if (sigaction(SIGFPE, &new, &old_sigact) == -1)
	{
		(void) fprintf(stderr,
			"sigaction(SIGFPE) failed in FPCOMP.h - errno %d\n",
			errno);
		return(-1);
	}

	return(0);
}

static void
restore_signal()
{
	(void) sigaction(SIGFPE, &old_sigact, (struct sigaction *)0);
}

static double
st(x)
double x;
{
	return(x);
}

static double
mlfabs(x)
double x;
{
	if (x < 0.0)
		return (-x);
	else
		return(x);
}

	/* determine which bits are used for exponent and sign */
	/* normalised numbers must have one of these bits set */
void
setnorm()
{
	double a = 1.0, b = -1.0, c = 2.0;
	char	*x;
	int	i;

	x = (char *) &a;
	for (i=0; i<sizeof(double); i++)
		normbits[i] = *x++;

	x = (char *) &b;
	for (i=0; i<sizeof(double); i++)
		normbits[i] |= *x++;

	x = (char *) &c;
	for (i=0; i<sizeof(double); i++)
		normbits[i] |= *x++;

	return;
}

	/* check whether a floating point number is normalised */
static int
testnorm(y)
double y;
{
	char 	*x;
	int	i;
	
	x = (char *) &y;
	
	for (i=0; i<sizeof(double); i++)
		if (normbits[i]&(*x++))
			return(0);
	return(1);
}

void
mlmachar()
{
	int	i, iz, k;
	int	irnd, negep;
	double	a, b, beta, betain, betam1, one, y, z, zero;
	float	c, d, e, f, fone, fzero;
	int	underflo ;
	int	normflag;
	char	*cp;
	volatile int	mx, inc;
	volatile double	fmax, newmax, mbase;
	static int	first_call = 1;

	if (!first_call)
		return;
	
	first_call = 0;

#ifdef SET_HWFACTOR
	/* this is only required in ".os" tests, and would cause
	   an unresolved reference to tet_getvar() in ".hdr" tests */

	if ((cp = tet_getvar("VSX_FP_SOFTWARE")) != 0 &&
		(*cp == 'Y' || *cp == 'y'))
	{
		hwfactor = 5;
	}
#endif

	one = 1.0;
	fone = 1.0;
	zero = 0.0;
	fzero = 0.0;

	/*   determine expbase,beta ala Malcolm	  */

	setnorm();
	a = one + one;
	while( st(st( st(a + one) - a) - one) == zero )
		a += a;
	b = one + one;
	while( st(st(a + b) - a) == zero )
		b += b;
	expbase = (int) ( st(a + b) - a);
	beta = expbase;

	/*   determine fsignif,dsignif,irnd   */
	fsignif = 0;
	c = fone;
	do {
		fsignif++;
		c *= beta;
		d = c + fone;
		e = d - c;
		f = e - fone;
	}
	while (f == fzero);

	dsignif = 0;
	b = one;
	do {
		dsignif++;
		b *= beta;
	}
	while ( st(st( st(b + one) - b) - one) == zero);

	irnd = 0;
	betam1 = beta - one;
	if( st(st((a + beta) + betam1) - (a + beta)) != zero )
		irnd = 1;

	/*   determine negep, epsneg   */

	negep = dsignif + 3;
	betain = one/beta;
	a = one;

	for( i = 1 ; i<=negep ; i++)
		a *= betain;

	b = a;
	while( st(st(one - a) - one) == zero )
	{
		a *= beta;
		negep--;
	};

	negep = - negep;
	epsneg = a;
	if( (expbase != 2) && (irnd != 0) )
	{
		a = a * st(one + a)/(one + one);
		if( st(st(one - a) - one) != zero )
			epsneg = a;
	};

	/*   determine epspos	*/

	a = b;
	while( st(st(one + a) - one) == zero )
	{
		a *= beta;
	}
	epspos = a;
	if( (expbase != 2) && (irnd != 0) )
	{
		a = a * st(one + a) / (one + one);
		if( st(st(one + a) - one) != zero )
			epspos = a;
	};

	/*  determine dexplen, dminexp, mindble

	      loop to determine largest i and k = 2**i such that
		  (beta) ** (2**(i))
	      does not underflow
	      exit from loop is signaled by an underflow   */

	i = 0;
	k = 1;
	z = betain;
	underflo = 0;
	while(!underflo)
	{
		y = z;
		z = y * y;

		/*   check for underflow   */
		/*   if this is not a normalised number consider as underflow */

		a = z * one;
		if( ( st(a + a) == zero) || (mlfabs(z) >=  y) )
			underflo = 1;
		else
		{
			if (expbase == 2 && testnorm(z))
			{
				normflag = 1;
				break;
			}
			i++;
			k += k;
		};
	}

	if( expbase != 10 )
	{
		dexplen = i + 1;
	}
	else
	{

		/*  for decimal machines only	*/
		dexplen = 2;
		iz = expbase;
		while( k >= iz )
		{
			iz *= expbase;
			dexplen = dexplen + 1;
		};
	}

	underflo = 0;
	while(!underflo)
	{

		/*   loop to determine dminexp, mindble
			  exit from loop is signalled by an underflow	 */

		mindble = y;
		y *= betain;
		/* check for underflow here */
		/*   if this is not a normalised number consider as underflow */
		a = y * one;
		if( ( st(a + a) == zero) || (mlfabs(y) >= mindble) ||
		    (st(y*st(one+epspos)) <= y) )
			underflo = 1;
		else
		{
			if (expbase == 2 && testnorm(y))
			{
				normflag = 1;
				break;
			}
			k++;
		}
	}
	dminexp = - k;

	/*  determine dmaxexp, maxdble	 */

	mx = 1;
	inc = 1;
	fmax = one;
	newmax = st(beta + one);
	mbase = beta;

	if (set_signal(sig_catch))
		return;

	/* Coarse Search */
	if (sigsetjmp(fpe_jmp, 1) == 0)
	{
		while (fmax < newmax)
		{
			fmax = newmax;
			newmax = st(newmax * mbase);

			/* Infinity reached? */
			if (newmax/mbase == newmax)
				break;
			mx += inc;
			inc++;
			mbase = st(mbase * beta);
		}
	}

	/* Fine search */
	newmax = fmax;
	fmax = one;
	if (sigsetjmp(fpe_jmp, 1) == 0)
	{
		while(fmax < newmax)
		{
			fmax = newmax;

			newmax = st(newmax * beta);

			/* Infinity reached? */
			if (newmax/beta == newmax)
				break;
			mx++;
		}
	}

	restore_signal();

	dmaxexp = ++mx;

	maxdble = one - epsneg;
	if( st(maxdble * one) != maxdble )
		maxdble = one - st(beta * epsneg);

	for (mx=0; mx < dmaxexp; mx++)
	{
		maxdble = st(maxdble * beta);
	}
	
	/* If denormalised values are possible adjust mindble and dminexp */
	if (normflag)
	{
		y = mindble;
		underflo = 0;
		while (!underflo)
		{
			mindble = y;
			y *= betain;
			a = st(y * one);
			if ( (st(a + a) == zero) || (mlfabs(y) >= mindble) )
				underflo++;
			else
				dminexp--;
		}
	}
}


fcomp(f1, f2)
float f1, f2;
{
	double	eps, pw, diff;
	double	f1a = f1, f2a = f2;
	int	i;
/* fcomp compares two single precision float numbers */

	mlmachar();
	i = 1;
	pw = 1.0;
lb:
	pw = pw / expbase;
	i = i + 1;
	if (i < (fsignif-1))       /* pw = power (expbase, 1.0 - dsignif) */
		goto lb;
	eps = 2 * pw / (1 - pw) * (1 - pw);

	if (f1 < 0.0)
		f1a = -f1;
	if (f2 < 0.0)
		f2a = -f2;

	if (f2 == 0.0 || (f2a < 1.0 && f1a > f2a * maxdble))
		diff = f1;
	else
		diff = 1.0 - (f1/f2);
	if (diff < 0.0)
		diff = -diff;
	if (diff <= eps)
		return 1;
	else
		return 0;
}

dcomp(d1, d2)
double d1, d2;
{
	double	eps, pw, diff;
	double	d1a = d1, d2a = d2;
	int	i;
/* dcomp compares two double precision float numbers */

	mlmachar();
	i = 1;
	pw = 1.0;
lb:
	pw = pw / expbase;
	i = i + 1;
	if (i < (dsignif - 2))       /* pw = power (expbase, 1.0 - dsignif) */
		goto lb;
	eps = 2 * pw / (1 - pw) * (1 - pw);

	if (d1 < 0.0)
		d1a = -d1;
	if (d2 < 0.0)
		d2a = -d2;

	if (d2 == 0.0 || (d2a < 1.0 && d1a > d2a * maxdble))
		diff = d1;
	else
		diff = 1.0 - (d1/d2);
	if (diff < 0.0)
		diff = -diff;
	if (diff <= eps)
		return 1;
	else
		return 0;
}
#if	LONG_DOUBLE_SUPPORTED
/*
 * Compare two long double numbers
 */
ldcomp(d1, d2)
long double d1, d2;
{
	long double	eps, pw, diff;
	long double	d1a = d1, d2a = d2;
	int	i;
/* dcomp compares two double precision float numbers */

	mlmachar();
	i = 1;
	pw = 1.0;
lb:
	pw = pw / expbase;
	i = i + 1;
	if (i < (dsignif - 2))       /* pw = power (B, 1.0 - dsignif) */
		goto lb;
	eps = 2 * pw / (1 - pw) * (1 - pw);

	if (d1 < 0.0)
		d1a = -d1;
	if (d2 < 0.0)
		d2a = -d2;

	if (d2 == 0.0 || (d2a < 1.0 && d1a > d2a * maxdble))
		diff = d1;
	else
		diff = 1.0 - (d1/d2);
	if (diff < 0.0)
		diff = -diff;
	if (diff <= eps)
		return(1);
	else
		return(0);
}
#endif /* LONG_DOUBLE_SUPPORTED */
